#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from typing import BinaryIO, Iterable
from mnp.mnp_url import MnpUrl
from mnp._stubs.mantik.mnp.mnp_pb2_grpc import MnpServiceStub
from mnp._stubs.mantik.mnp.mnp_pb2 import PushRequest

import grpc
from io import DEFAULT_BUFFER_SIZE


class Forwarder:
    """
    Simple thread based forwarding routine which takes some input and forwards it
    to a nother MNP Node
    """

    def __init__(self, task_id: str, binary_input: BinaryIO, destination_url: str):
        self.binary_input = binary_input
        self.task_id = task_id
        self.destination = MnpUrl.parse(destination_url)

    def run(self):
        channel = grpc.insecure_channel(self.destination.address)
        try:
            self.run_with_channel(channel)
        finally:
            channel.close()

    def run_with_channel(self, channel: grpc.Channel):
        stub = MnpServiceStub(channel)

        def creator() -> Iterable[PushRequest]:
            yield PushRequest(
                session_id=self.destination.session_id,
                task_id=self.task_id,
                port=self.destination.port
            )
            data = self.binary_input.read(DEFAULT_BUFFER_SIZE)
            while data:
                yield PushRequest(data=data)
                data = self.binary_input.read(DEFAULT_BUFFER_SIZE)
            yield PushRequest(
                done=True
            )

        stub.Push(creator())



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from __future__ import annotations
from urllib.parse import urlparse

class MnpUrl:

    def __init__(self, address: str, session_id: str, port: int):
        self.address = address
        self.session_id = session_id
        self.port = port

    @classmethod
    def parse(cls, s: str) -> MnpUrl:
        try:
            parsed = urlparse(s)
            if parsed.scheme != "mnp":
                raise ValueError("Expected scheme mnp")
            address = "{}:{}".format(parsed.hostname, parsed.port)
            elements = list(filter(None, parsed.path.split("/")))
            session_id = elements[0]
            port = int(elements[1])
            return cls(address, session_id, port)
        except Exception as e:
            raise ValueError("Invalid url", e)

    def format(self) -> str:
        return "mnp://{}/{}/{}".format(self.address, self.session_id, self.port)



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from mnp.service_server import ServiceServer
from mnp.handler import Handler
from mnp.session_handler import SessionHandler
from mnp.server import Server
from mnp.elements import *
from mnp.mnp_url import MnpUrl



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

import abc
from google.protobuf.any_pb2 import Any
from mnp.session_handler import SessionHandler
from mnp.elements import *
from typing import Callable, List, Iterator


class Handler(metaclass=abc.ABCMeta):
    """
    A MNP Handler
    """

    @abc.abstractmethod
    def about(self) -> AboutResponse:
        """Responses to About call"""

    @abc.abstractmethod
    def quit(self):
        """Quit requested"""

    @abc.abstractmethod
    def init_session(self, session_id: str,
                     configuration: Any,
                     ports: PortConfiguration,
                     callback: Callable[[SessionState], None] = None) -> SessionHandler:
        """Initialize a session"""



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

import abc
from mnp.elements import PortConfiguration, OutputPortConfiguration
from mnp.collectable_byte_buffer import CollectableByteBuffer
from typing import List, BinaryIO, cast
import io


class SessionHandler(metaclass=abc.ABCMeta):
    """
    MNP SessionHandler
    """

    def __init__(self, session_id: str, ports: PortConfiguration):
        self.session_id = session_id
        self.ports = ports

    @abc.abstractmethod
    def quit(self):
        """
        Quit the session
        """

    @abc.abstractmethod
    def run_task(self, task_id: str, inputs: List[BinaryIO], outputs: List[BinaryIO]):
        """
        run a task
        """

    def run_task_with_bytes(self, task_id: str, inputs: List[bytes]) -> List[bytes]:
        """
        run a task with in memory byte representation
        (convenience method)
        """

        def open_reader(b: bytes) -> BinaryIO:
            return io.BytesIO(b)

        input_readers = list(map(open_reader, inputs))

        def open_writer(content_type: OutputPortConfiguration) -> CollectableByteBuffer:
            return CollectableByteBuffer()

        output_writers = list(map(open_writer, self.ports.outputs))
        self.run_task(task_id, input_readers, cast(List[BinaryIO], output_writers))

        def read_result(writer: CollectableByteBuffer) -> bytes:
            return writer.fetch_collected_bytes()

        results = list(map(read_result, output_writers))
        return results



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

import grpc
from concurrent import futures
from mnp import ServiceServer, Handler
from mnp._stubs.mantik.mnp.mnp_pb2_grpc import add_MnpServiceServicer_to_server


class Server:

    def __init__(self, handler: Handler, tpe: futures.ThreadPoolExecutor = None):
        if tpe is None:
            tpe = futures.ThreadPoolExecutor(max_workers=10)
        self.handler = handler
        self.service = ServiceServer(handler, tpe)
        self.gprc_server = grpc.server(tpe)
        add_MnpServiceServicer_to_server(self.service, self.gprc_server)

    def start(self, address: str):
        self.gprc_server.add_insecure_port(address)
        self.gprc_server.start()

    def wait_for_termination(self):
        self.gprc_server.wait_for_termination()

    def stop(self, grace=None):
        self.gprc_server.stop(grace)



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from io import BytesIO

class CollectableByteBuffer(BytesIO):
    """
    A Byte buffer which we can collect after it has been closed.
    (Technically it just disables closing)
    """

    def __init__(self) -> None:
        super().__init__(b"")
        self.close_called = False

    def close(self) -> None:
        # Do not close, otherwise we can't read bytes anymore
        self.close_called = True

    def fetch_collected_bytes(self) -> bytes:
        self.seek(0)
        b = self.read()
        super().close()
        return b





#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from typing import List
from typing import BinaryIO
from mnp._stubs.mantik.mnp.mnp_pb2 import TaskPortStatus
import io

import os

# Max Chunk size to aovid deadlock via Pipes
MAX_CHUNK_SIZE = 8192


class StreamyPipe:

    def __init__(self):
        read_fd, write_fd = os.pipe()
        self.read_fd = read_fd
        self.write_fd = write_fd
        self.writer = io.open(write_fd, "wb")
        self.reader = io.open(read_fd, "rb")
        self.calls = 0
        self.byte_count = 0
        self.is_done = False

    def write(self, b: bytes):
        """
        Write something on the incoming side
        """
        result = self.writer.write(b)
        self.calls += 1
        self.byte_count += len(b)
        return result

    def write_eof(self):
        """
        Close incoming side
        """
        result = self.writer.close()
        self.calls += 1
        self.is_done = True
        return result

    def read(self) -> bytes:
        """
        Read out side
        """
        result = self.reader.read(MAX_CHUNK_SIZE)
        self.calls += 1
        self.byte_count += len(result)
        if not result:
            self.is_done = True
        return result

    def task_port_status(self) -> TaskPortStatus:
        # No error support here
        return TaskPortStatus(
            msg_count=self.calls,
            data=self.byte_count,
            done=self.is_done
        )


class StreamMultiplexer:
    """
    Handles input and output streams of a Task
    """

    inputs: List[StreamyPipe]
    outputs: List[StreamyPipe]

    def __init__(self, input_num: int, output_num: int):
        self.inputs = [StreamyPipe() for i in range(input_num)]
        self.outputs = [StreamyPipe() for i in range(output_num)]

    def wrapped_inputs(self) -> List[BinaryIO]:
        return [i.reader for i in self.inputs]

    def wrapped_outputs(self) -> List[BinaryIO]:
        return [i.writer for i in self.outputs]

    def push(self, port: int, b: bytes):
        return self.inputs[port].write(b)

    def push_eof(self, port: int):
        return self.inputs[port].write_eof()

    def pull(self, port: int) -> bytes:
        return self.outputs[port].read()

    def output_reader(self, port: int) -> BinaryIO:
        return self.outputs[port].reader

    def close_output(self):
        for output in self.outputs:
            output.writer.close()

    def close(self):
        for output in self.outputs:
            output.writer.close()
        for input in self.inputs:
            input.writer.close()

    def input_port_status(self) -> List[TaskPortStatus]:
        """
        Reports input port status in an MNP Compatible way
        """
        return list(map(StreamyPipe.task_port_status, self.inputs))

    def output_port_status(self) -> List[TaskPortStatus]:
        """
        Report output port status in a MNP Compatible way.
        """
        return list(map(StreamyPipe.task_port_status, self.outputs))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mnp']

package_data = \
{'': ['*']}

install_requires = \
['grpcio-tools>=1.37.1,<2.0.0',
 'grpcio>=1.37.1,<2.0.0',
 'protobuf>=3.16.0,<4.0.0']

setup_kwargs = {
    'name': 'mnp',
    'version': '0.3.0rc4',
    'description': '',
    'long_description': None,
    'author': 'Mantik Team',
    'author_email': 'info@mantik.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.mantik.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.8',
}


setup(**setup_kwargs)



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from typing import List, Optional
from dataclasses import dataclass
from google.protobuf.any_pb2 import Any
from enum import IntEnum
from mnp._stubs.mantik.mnp.mnp_pb2 import SS_INITIALIZING, SS_DOWNLOADING, SS_STARTING_UP, SS_READY, SS_FAILED
from mnp._stubs.mantik.mnp.mnp_pb2 import AboutResponse as ProtoAboutResponse, \
    ConfigureInputPort as ProtoConfigureInputPort, ConfigureOutputPort as ProtoConfigureOutputPort, \
    InitRequest as ProtoInitRequest
from mnp.rpc_converter import string_or_none, or_empty_string


# Nice python wrappers for MNP Elements

@dataclass
class AboutResponse:
    name: str
    extra: Optional[Any] = None

    def as_proto(self) -> ProtoAboutResponse:
        return ProtoAboutResponse(
            name=self.name,
            extra=self.extra
        )


@dataclass
class InputPortConfiguration:
    content_type: str

    @classmethod
    def from_proto(cls, p: ProtoConfigureInputPort):
        return cls(content_type=p.content_type)

    def as_proto(self) -> ProtoConfigureInputPort:
        return ProtoConfigureInputPort(content_type=self.content_type)


@dataclass
class OutputPortConfiguration:
    content_type: str
    forwarding: Optional[str] = None

    @classmethod
    def from_proto(cls, p: ProtoConfigureOutputPort):
        return cls(
            p.content_type,
            string_or_none(p.destination_url)
        )

    def as_proto(self) -> ProtoConfigureOutputPort:
        return ProtoConfigureOutputPort(
            content_type=self.content_type,
            destination_url=or_empty_string(self.forwarding)
        )


@dataclass
class PortConfiguration:
    inputs: List[InputPortConfiguration]
    outputs: List[OutputPortConfiguration]

    @classmethod
    def from_init(cls, init: ProtoInitRequest):
        return cls(
            list(map(InputPortConfiguration.from_proto, init.inputs)),
            list(map(OutputPortConfiguration.from_proto, init.outputs))
        )

    def inputs_as_proto(self) -> List[ProtoConfigureInputPort]:
        return list(map(InputPortConfiguration.as_proto, self.inputs))

    def outputs_as_proto(self) -> List[ProtoConfigureOutputPort]:
        return list(map(OutputPortConfiguration.as_proto, self.outputs))


class SessionState(IntEnum):
    INITIALIZING = SS_INITIALIZING
    DOWNLOADING = SS_DOWNLOADING
    STARTING_UP = SS_STARTING_UP
    READY = SS_READY
    FAILED = SS_FAILED



#
# This file is part of the Mantik Project.
# Copyright (c) 2020-2021 Mantik UG (Haftungsbeschränkt)
# Authors: See AUTHORS file
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
#
# Additionally, the following linking exception is granted:
#
# If you modify this Program, or any covered work, by linking or
# combining it with other code, such other code is not for that reason
# alone subject to any of the requirements of the GNU Affero GPL
# version 3.
#
# You can be released from the requirements of the license by purchasing
# a commercial license.
#

from concurrent import futures
from queue import Queue
from typing import Dict, Iterable, Optional

from mnp._stubs.mantik.mnp.mnp_pb2 import QuitResponse, InitResponse, SS_READY, SS_FAILED, PushResponse, PullResponse, \
    QueryTaskResponse, TS_UNKNOWN, TS_EXISTS, TS_FINISHED, TS_FAILED
from mnp._stubs.mantik.mnp.mnp_pb2 import AboutResponse as MnpAboutResponse
from mnp._stubs.mantik.mnp.mnp_pb2_grpc import MnpServiceServicer
from mnp.elements import SessionState as ESessionState, PortConfiguration
from mnp.handler import Handler, SessionHandler
from mnp.streammultiplexer import StreamMultiplexer
from mnp.forwarder import Forwarder
from threading import Lock
import logging


class SessionTask:
    """
    Handles a task on the server
    """

    def __init__(self, task_id: str, port_config: PortConfiguration):
        self.task_id = task_id
        self.port_config = port_config
        self.multiplexer = StreamMultiplexer(len(port_config.inputs), len(port_config.outputs))
        self.error_message = ""
        self.state = TS_EXISTS

    def start(self, executor: futures.Executor, session_handler: SessionHandler):
        """
        Start this task, it's running asynchronous
        """
        logging.info("Session %s Starting Task %s", session_handler.session_id, self.task_id)

        inputs = self.multiplexer.wrapped_inputs()
        outputs = self.multiplexer.wrapped_outputs()
        future = executor.submit(session_handler.run_task, self.task_id, inputs, outputs)
        self.start_portforwarders(executor)

        def on_done(future):
            try:
                future.result()
                logging.debug("Session %s Task %s done", session_handler.session_id, self.task_id)
                self.state = TS_FINISHED
            except Exception as e:
                msg = repr(e)
                logging.warning("Task %s failed, %s", self.task_id, msg, exc_info=True)
                self.error(repr(e))
            finally:
                self.multiplexer.close_output()

        future.add_done_callback(on_done)

    def start_portforwarders(self, executor: futures.Executor):
        started = 0
        for idx, output in enumerate(self.port_config.outputs):
            if output.forwarding is not None:
                forwarder = Forwarder(self.task_id, self.multiplexer.output_reader(idx), output.forwarding)
                executor.submit(forwarder.run)
                started += 1
        if started > 0:
            logging.info("Started %d forwarders", started)

    def error(self, error: str):
        """
        Set this task into an error state
        """
        logging.error("Task Error %s", error)
        self.error_message = error
        self.multiplexer.close()
        if self.state != TS_FINISHED:
            self.state = TS_FAILED
        pass

    def query(self) -> QueryTaskResponse:
        return QueryTaskResponse(
            state=self.state,
            error=self.error_message,
            inputs=self.multiplexer.input_port_status(),
            outputs=self.multiplexer.output_port_status()
        )


class SessionServer:
    """
    Handles a session on the server
    """

    tasks: Dict[str, SessionTask]

    def __init__(self, session_id: str, session_handler: SessionHandler, executor: futures.Executor):
        self.session_id = session_id
        self.handler = session_handler
        self.mutex = Lock()
        self.tasks = {}
        self.executor = executor

    def get_or_create_task(self, task_id: str) -> SessionTask:
        self.mutex.acquire()
        try:
            task = self.tasks.get(task_id)
            if task is None:
                task = SessionTask(task_id, self.handler.ports)
                self.tasks[task_id] = task
                task.start(self.executor, self.handler)
            return task
        except Exception as e:
            logging.error("Could not acquire task {}".format(task_id), exc_info=True)
            raise e
        finally:
            self.mutex.release()

    def get_task(self, task_id: str) -> Optional[SessionTask]:
        return self.tasks.get(task_id)

    def close(self):
        self.handler.quit()
        # No way yet to close running tasks, but we can stop their channels
        for task_id, task in self.tasks.items():
            task.error("Shutdown")


class ServiceServer(MnpServiceServicer):
    sessions: Dict[str, SessionServer]

    def __init__(self, handler: Handler, executor: futures.Executor):
        self.handler = handler
        self.sessions = {}
        self.executor = executor

    def About(self, request, context):
        response = self.handler.about()
        return response.as_proto()

    def Quit(self, request, context):
        self.handler.quit()
        for session_id, session in self.sessions.items():
            session.close()
        return QuitResponse()

    def Init(self, request, context):
        # Note: this methods returns an Iterable for the responses!

        ports = PortConfiguration.from_init(request)
        session_id = request.session_id
        configuration = request.configuration

        # Queue is not iterable, but we can convert to it, see https://stackoverflow.com/a/11288503
        sentinel = object()

        state_queue = Queue()

        def callback(state: ESessionState) -> None:
            logging.info("Session state %s", state)
            if state != ESessionState.READY and state != ESessionState.FAILED:
                state_queue.put(
                    InitResponse(
                        state=state.value
                    )
                )

        session_handler_future = self.executor.submit(
            self.handler.init_session,
            session_id,
            configuration,
            ports,
            callback
        )

        def done(f: futures.Future):
            try:
                session_handler = f.result()
                self.sessions[session_id] = SessionServer(session_id, session_handler, self.executor)
                logging.info("Initialized session %s", session_id)
                state_queue.put(
                    InitResponse(
                        state=SS_READY
                    )
                )
                state_queue.put(sentinel)
            except Exception as e:
                error_message = repr(e)
                logging.warning("Could not start session %s", session_id, exc_info=True)
                state_queue.put(
                    InitResponse(
                        state=SS_FAILED,
                        error=error_message
                    )
                )
                state_queue.put(sentinel)

        session_handler_future.add_done_callback(done)

        return iter(state_queue.get, sentinel)

    def Push(self, request_iterator, context):
        first = next(request_iterator)
        session = self.sessions.get(first.session_id)

        if session is None:
            context.set_details("Unknown session")
            raise ValueError("Unknown session")
        task_id = first.task_id
        port = first.port
        logging.debug("Pushing %s/%s/%d", first.session_id, task_id, port)
        task = session.get_or_create_task(task_id)

        task.multiplexer.push(port, first.data)
        done = first.done
        for n in request_iterator:
            task.multiplexer.push(port, n.data)
            done = n.done

        if done:
            task.multiplexer.push_eof(port)
        else:
            task.error("No EOF on port {}".format(port))

        return PushResponse()

    def Pull(self, request, context):
        session = self.sessions.get(request.session_id)
        if session is None:
            context.set_details("Unknown session")
            raise ValueError("Unknown session")

        task_id = request.task_id
        port = request.port
        logging.debug("Pulling %s/%s/%d", request.session_id, task_id, port)
        task = session.get_or_create_task(task_id)

        def respond() -> Iterable[PullResponse]:
            while True:
                data = task.multiplexer.pull(port)
                if not data: break
                yield PullResponse(data=data)
            yield PullResponse(done=True)

        return respond()

    def QuitSession(self, request, context):
        session = self.sessions.get(request.session_id)
        if session is None:
            context.set_details("Unknown session")
            raise ValueError("Unknown session")
        session.close()
        self.sessions.pop(request.session_id, None)
        return QuitResponse()

    def QueryTask(self, request, context):
        session = self.sessions.get(request.session_id)
        if session is None:
            context.set_details("Unknown session")
            raise ValueError("Unknown session")
        task_id = request.task_id
        ensure: bool = request.ensure
        task: SessionTask
        if ensure:
            task = session.get_or_create_task(task_id)
        else:
            task = session.get_task(task_id)

        if task is None:
            return QueryTaskResponse(state=TS_UNKNOWN)

        return task.query()
